#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cerrno>

#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <string>

using namespace std;

#include "core.hh"

extern "C" {
#include "OpenSPC.h"
}


static const unsigned core_rate = 32000;

static FILE* resampler = NULL;
static unsigned Bufsize    = 735*4;

static unsigned Rate       = 44100;
static unsigned Channels   = 2;
static unsigned Bitness    = 16;
static bool Mute = false;
//static bool Alt  = false;
static bool Echo = true;
static bool Interpolation = true;
static int cur_rate = 100;
static unsigned chan_flags = 0xFF;

static void CloseResampler()
{
    if(resampler) { pclose(resampler); resampler = NULL; }
}
static void OpenResampler()
{
    int pip[2];
    socketpair(AF_UNIX, SOCK_STREAM, 0, pip);
    if(!fork())
    {
        char Buf[512];
        unsigned InputRate = core_rate;
        
        std::string inputspec;
        sprintf(Buf, "r%u", InputRate);
        inputspec += Buf;


        std::string outputspec;
        if(Channels == 1) outputspec += 'm';
        if(Bitness  == 8) outputspec += 'b';
        sprintf(Buf, "r%u", Rate);
        outputspec += Buf;
        
        int p0 = dup(dup(pip[0]));
        dup2(p0, 0);
        dup2(p0, 1);
        for(int n=3; n<511; ++n) close(n);
        execlp("resample",
               "resample",
               inputspec.c_str(),
               outputspec.c_str(),
               Interpolation ? "-flanczos" : "-flinear",
               NULL);
        
        perror("execlp(\"resample\")");
        _exit(-1);
    }
    close(pip[0]);
    
    resampler = fdopen(pip[1], "r+");
}


void SPCCore_MuteSound()        { Mute = true; }
void SPCCore_SetChannels(int n) { CloseResampler(); Channels = n; }
void SPCCore_SetBitness(int n)  { CloseResampler(); Bitness = n; }
void SPCCore_SetInterpolation(int n) { CloseResampler(); Interpolation = n; }
void SPCCore_SetEcho(int n) { Echo = n; }
void SPCCore_SetAlt(int n)  { }
void SPCCore_SetRate(int n) { CloseResampler(); Rate = n; }
void SPCCore_SetPlaybackSpeed(int n)
{
    if(n < 100) n = 100;
    cur_rate = n;
    SNDurate = cur_rate;
    SNDmixlen = SNDfreq / SNDurate;
}
void SPCCore_SetChannelStatus(int n, int s)
{
    if(s)
        chan_flags |= (1<<n);
    else
        chan_flags &= ~(1<<n);
}

int SPCCore_GetAlt()  { return 0;/*Alt*/ }
int SPCCore_GetEcho() { return Echo; }
int SPCCore_GetRate() { return Rate; }
int SPCCore_GetChannels() { return Channels; }
int SPCCore_GetBitness() { return Bitness; }
int SPCCore_GetChannelStatus(int n) { return chan_flags&(1<<n); }
int SPCCore_GetPlaybackSpeed() { return cur_rate; }
int SPCCore_GetInterpolation() { return Interpolation; }

void SPCCore_Terminate()
{
    CloseResampler();
}

unsigned SPCCore_Init()
{
    return Bufsize;
}
int SPCCore_Load(const char* fn)
{
    int res = SPCInit(const_cast<char*> (fn));
    printf("Load result: %d\n", res);

    SNDInit(core_rate/*SPCCore_GetRate()*/, 100, 1, SPCCore_GetChannels()==2, SPCCore_GetBitness()==8);
    SNDvmask = 0xFF;
    SNDNoteOn(SPC_DSP_Buffer[0x4c]);
    SNDcrc=0;
    
    return res ? -1 : 0;
}

void SPCCore_Generate(void* buffer)
{
    if(!resampler) OpenResampler();
    
    SNDmix=1;
    
    int fd = fileno(resampler);
    
    unsigned left = Bufsize;
    unsigned char *dest = (unsigned char *) buffer;
    while(left > 0)
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
        
        int n = read(fd, dest, left);
        if(!n)
        {
            /* EOF */
            break;
        }
        if(n < 0)
        {
            if(errno != EAGAIN) break;
            if(errno == EINTR) continue;

            /* needs more data */
            
            fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
        
            SNDMix();
            if(SNDmix)
            {
              unsigned mul = Channels * (Bitness/8);
              //printf("Mixing %d * %u\n", SNDmixlen, mul);
              write(fd, SNDoptbuf, SNDmixlen * mul);
            }
            
            continue;
        }
        
        left -= n;
        dest += n;
    }
}

SPCCore_ID* SPCCore_GetID(const char *fn)
{
    static SPCCore_ID stat;
    memset(&stat, 0, sizeof(stat));
    
    if(SPCname) strncpy(stat.songname,  SPCname,  33);
    if(SPCtname) strncpy(stat.gametitle, SPCtname, 33);
    if(SPCdname)   strncpy(stat.dumper,    SPCdname,    17);
    if(SPCcomment) strncpy(stat.comments,  SPCcomment,  33);
    if(SPCdate)   strncpy(stat.author,    SPCdate,    33);
    
    stat.playtime = SPCtime;
    stat.fadetime = 0;
    stat.emulator = 0;
    return &stat;
}

extern "C" void Shutdown() { }
