#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cerrno>

#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <string>

using namespace std;

#include "core.hh"

extern "C" {
#include "openspc.h"
#include "id666.h"
}


static FILE* resampler = NULL;
static unsigned Bufsize    = 1024;

static unsigned Rate       = 32000;
static unsigned Channels   = 2;
static unsigned Bitness    = 16;
static bool Mute = false;
//static bool Alt  = false;
static bool Echo = true;
static bool Interpolation = true;
static int cur_rate = 100;
static unsigned chan_flags = 0xFF;


static void CloseResampler()
{
    if(resampler) { pclose(resampler); resampler = NULL; }
}
static void OpenResampler()
{
    int pip[2];
    socketpair(AF_UNIX, SOCK_STREAM, 0, pip);
    if(!fork())
    {
        char Buf[512];
        unsigned InputRate = 32000;
        
        std::string inputspec;
        sprintf(Buf, "r%u", InputRate);
        inputspec += Buf;


        std::string outputspec;
        if(Channels == 1) outputspec += 'm';
        if(Bitness  == 8) outputspec += 'b';
        sprintf(Buf, "r%u", Rate);
        outputspec += Buf;
        
        int p0 = dup(dup(pip[0]));
        dup2(p0, 0);
        dup2(p0, 1);
        for(int n=3; n<511; ++n) close(n);
        execlp("resample",
               "resample",
               inputspec.c_str(),
               outputspec.c_str(),
               Interpolation ? "-flanczos" : "-flinear",
               NULL);
        
        perror("execlp(\"resample\")");
        _exit(-1);
    }
    close(pip[0]);
    
    resampler = fdopen(pip[1], "r+");
}


void SPCCore_MuteSound()        { Mute = true; }
void SPCCore_SetChannels(int n) { CloseResampler(); Channels = n; }
void SPCCore_SetBitness(int n)  { CloseResampler(); Bitness = n; }
void SPCCore_SetInterpolation(int n) { CloseResampler(); Interpolation = n; }
void SPCCore_SetEcho(int n) { Echo = n; }
void SPCCore_SetAlt(int n)  { }
void SPCCore_SetRate(int n) { CloseResampler(); Rate = n; }
void SPCCore_SetPlaybackSpeed(int n) { cur_rate = n; }
void SPCCore_SetChannelStatus(int n, int s)
{
    if(s)
        chan_flags |= (1<<n);
    else
        chan_flags &= ~(1<<n);
}

int SPCCore_GetAlt()  { return 0;/*Alt*/ }
int SPCCore_GetEcho() { return Echo; }
int SPCCore_GetRate() { return Rate; }
int SPCCore_GetChannels() { return Channels; }
int SPCCore_GetBitness() { return Bitness; }
int SPCCore_GetChannelStatus(int n) { return chan_flags&(1<<n); }
int SPCCore_GetPlaybackSpeed() { return cur_rate; }
int SPCCore_GetInterpolation() { return Interpolation; }

void SPCCore_Terminate()
{
    CloseResampler();
}

unsigned SPCCore_Init()
{
    return Bufsize;
}
int SPCCore_Load(const char* fn)
{
    FILE* fp = fopen(fn, "rb");
    if(!fp) return -1;

    fseek(fp, 0, SEEK_END);
    
    unsigned size = ftell(fp);
    
    char *buf = new char[size];
    
    rewind(fp);
    fread(buf, 1, size, fp);
    fclose(fp);
    
    int ret = OSPC_Init(buf, size);
    
    delete[] (buf);
    
    if(ret==0) return 0; /* ok */
    if(ret > 1) return -1; /* file type failure */
    return -1;
}

void SPCCore_Generate(void* buffer)
{
    if(!resampler) OpenResampler();
    
    int fd = fileno(resampler);
    
    unsigned left = Bufsize;
    unsigned char *dest = (unsigned char *) buffer;
    while(left > 0)
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
        
        int n = read(fd, dest, left);
        if(!n)
        {
            /* EOF */
            break;
        }
        if(n < 0)
        {
            if(errno != EAGAIN) break;
            if(errno == EINTR) continue;

            /* needs more data */
            
            fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
        
            char tmp[4096];
            unsigned bytes = OSPC_Run(-1, (short *)tmp, sizeof(tmp));
            
            write(fd, tmp, bytes);
            
            continue;
        }
        
        left -= n;
        dest += n;
    }
}

SPCCore_ID* SPCCore_GetID(const char *fn)
{
    int fd = open(fn, O_RDONLY);
    if(fd < 0) return NULL;
    
    id666_t tmp;
    
    if(spc_read_id666(fd, &tmp) < 0)
    {
        close(fd);
        return NULL;
    }
    
    close(fd);
    
    static SPCCore_ID stat;
    memset(&stat, 0, sizeof(stat));
    
    if(tmp.song_title) strncpy(stat.songname,  tmp.song_title,  33);
    if(tmp.game_title) strncpy(stat.gametitle, tmp.game_title, 33);
    if(tmp.dumper)   strncpy(stat.dumper,    tmp.dumper,    17);
    if(tmp.comments) strncpy(stat.comments,  tmp.comments,  33);
    if(tmp.artist)   strncpy(stat.author,    tmp.artist,    33);

    stat.playtime = tmp.loop_length + tmp.intro_length;
    stat.fadetime = tmp.fade_length;
    
    if(!strcmp(tmp.emulator, "zsnes"))
       stat.emulator = 1;
    else if(!strcmp(tmp.emulator, "snes9x"))
       stat.emulator = 2;
    else
       stat.emulator = strtol(tmp.emulator, NULL, 10);
    
    return &stat;
}
